/*
    LWINS - Lightweight WINS Server http://www.lwins.org/

    Copyright (c) 2008 - 2009 by Thomas Bluemel <thomasb@lwins.org>

    This file is part of LWINS.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
#include <lwins.h>

static const unsigned int g_loglevel = 0xffffffff;
static char g_szReqId[4] = {0};

void lwins_assert (const char *msg, const char *file, int line)
{
    fprintf (stderr, "ASSERTION FAILED: %s at %s:%d\n", msg, file, line);
    exit (1);
}

void lwins_dbg_start_nbns (u16 tran_id)
{
    sprintf (g_szReqId, "N%04x", tran_id & 0xFFFF);
}

void lwins_dbg_start_discover (void)
{
    static unsigned int req = 0;
    
    sprintf (g_szReqId, "D%04x", req);
    if (++req > 0xFFFF)
        req = 0;
}

void lwins_dbg_start_assoc (u32 handle)
{
    if (handle > 0xFFFF)
        strcpy (g_szReqId, "A????");
    else
        sprintf (g_szReqId, "A%04x", handle);
}

void lwins_dbg_start (void)
{
    memset (g_szReqId, 0, sizeof (g_szReqId));
}

void lwins_log (int level, const char *file, int line, const char *fmt, ...)
{
    va_list args;
    struct tm *ptm;
    struct timeval tv;
    
    (void)file;
    (void)line;
    
    if (gettimeofday (&tv, NULL) == 0)
    {
        ptm = localtime (&tv.tv_sec);
        fprintf (stderr, "%02d:%02d:%02d.%03ld ", ptm->tm_hour, ptm->tm_min, ptm->tm_sec, tv.tv_usec / 1000);
    }
    
    fprintf (stderr, "%4s ", g_szReqId);
    va_start (args, fmt);
    if (!level || (g_loglevel & (1 << level)))
        vfprintf (stderr, fmt, args);
    va_end (args);
}

int lwins_check_loglevel (int level)
{
    if (!level || (g_loglevel & (1 << level)))
        return 1;
    return 0;
}

static void lwins_stack_trace (int signum, siginfo_t *info, void *ptr)
{
    size_t addr_cnt;
    
    (void)signum;
    (void)info;
    (void)ptr;
    
    //addr_cnt = backtrace (addrs, addrs_max);
    if (addr_cnt != 0)
        //backtrace_symbols_fd (addrs, addr_cnt, fileno (stderr));
    abort ();
}

void lwins_setup_stack_traces (void)
{
    struct sigaction action;
    
    memset (&action, 0, sizeof (action));
    action.sa_sigaction = lwins_stack_trace;
    action.sa_flags = SA_SIGINFO | SA_ONSTACK;
    sigaction (SIGSEGV, &action, NULL);
}

static void lwins_dump_buffer_hex (const unsigned char *buffer, size_t len)
{
    char line[(16 * 3) + 2 + 16 + 2];
    size_t i, j, n;
    char hex[3];
    
    line[sizeof (line) - 2] = '\n';
    line[sizeof (line) - 1] = '\0';
    
    i = 0;
    while (i < len)
    {
        memset (line, ' ', sizeof (line) - 2);
        n = len - i;
        if (n > 16)
            n = 16;
        
        for (j = 0; j < n; j++)
        {
            sprintf (hex, "%02X", buffer[i + j]);
            line[j * 3] = hex[0];
            line[(j * 3) + 1] = hex[1];
            line[(16 * 3) + 2 + j] = (isprint ((char)buffer[i + j]) ? buffer[i + j] : '.');
        }
        
        LOG (LOG_PACKET, "%s", line);
        i += n;
    }
}

static void lwins_dump_resource_record (const char *name, struct netbt_resource *resource, unsigned int idx)
{
    LOG (LOG_PACKET, "--- %s[%u].rname: %s\n", name, idx, resource[idx].rname);
    LOG (LOG_PACKET, "--- %s[%u].rtype: %u\n", name, idx, resource[idx].rtype);
    LOG (LOG_PACKET, "--- %s[%u].rclass: %u\n", name, idx, resource[idx].rclass);
    LOG (LOG_PACKET, "--- %s[%u].rttl: %u\n", name, idx, resource[idx].rttl);
    LOG (LOG_PACKET, "--- %s[%u].rdlen: %u\n", name, idx, resource[idx].rdlen);
    lwins_dump_buffer_hex (resource[idx].rdata, resource[idx].rdlen);
}

void lwins_dump_nbns_raw_packet (unsigned char *raw_packet, size_t len, struct sockaddr_in *client)
{
    u16 tmpu16;
    struct netbt_packet packet;
    
    LOG (LOG_PACKET, "=== PACKET FROM %s:%u SIZE: %zd ===\n", inet_ntoa (client->sin_addr), ntohs(client->sin_port), len);
    
    if (netbt_parse_raw_packet (raw_packet, len, client, &packet))
    {
        LOG (LOG_PACKET, "- HEADER\n");
        LOG (LOG_PACKET, "--- tran_id: %u\n", packet.header.tran_id);
        LOG (LOG_PACKET, "--- is_response: %u\n", packet.header.is_response);
        LOG (LOG_PACKET, "--- op_code: %u\n", packet.header.op_code);
        LOG (LOG_PACKET, "--- authoritive_answer: %u\n", packet.header.authoritive_answer);
        LOG (LOG_PACKET, "--- truncation: %u\n", packet.header.truncation);
        LOG (LOG_PACKET, "--- recursion_desired: %u\n", packet.header.recursion_desired);
        LOG (LOG_PACKET, "--- recursion_available: %u\n", packet.header.recursion_available);
        LOG (LOG_PACKET, "--- broadcast: %u\n", packet.header.broadcast);
        LOG (LOG_PACKET, "--- response_code: %u\n", packet.header.response_code);
        LOG (LOG_PACKET, "--- qscnt: %u\n", packet.header.qscnt);
        LOG (LOG_PACKET, "--- ancnt: %u\n", packet.header.ancnt);
        LOG (LOG_PACKET, "--- nscnt: %u\n", packet.header.nscnt);
        LOG (LOG_PACKET, "--- arcnt: %u\n", packet.header.arcnt);
        if (packet.header.qscnt > 0)
            LOG (LOG_PACKET, "- QUESTIONS\n");
        for (tmpu16 = 0; tmpu16 < packet.header.qscnt; tmpu16++)
        {
            LOG (LOG_PACKET, "--- questions[%u].qname: %s\n", tmpu16, packet.questions[tmpu16].qname);
            LOG (LOG_PACKET, "--- questions[%u].qtype: %u\n", tmpu16, packet.questions[tmpu16].qtype);
            LOG (LOG_PACKET, "--- questions[%u].qclass: %u\n", tmpu16, packet.questions[tmpu16].qclass);
        }
        if (packet.header.ancnt > 0)
            LOG (LOG_PACKET, "- ANSWERS\n");
        for (tmpu16 = 0; tmpu16 < packet.header.ancnt; tmpu16++)
            lwins_dump_resource_record ("answers", packet.answers, tmpu16);
        if (packet.header.nscnt > 0)
            LOG (LOG_PACKET, "- AUTHORITIES\n");
        for (tmpu16 = 0; tmpu16 < packet.header.nscnt; tmpu16++)
            lwins_dump_resource_record ("authority", packet.authority, tmpu16);
        if (packet.header.arcnt > 0)
            LOG (LOG_PACKET, "- ADDITIONAL\n");
        for (tmpu16 = 0; tmpu16 < packet.header.arcnt; tmpu16++)
            lwins_dump_resource_record ("additional", packet.additional, tmpu16);
        netbt_free_packet (&packet);
    }
    else
        LOG (LOG_PACKET, "Could not parse packet\n");
    
    lwins_dump_buffer_hex (raw_packet, len);
    
    LOG (LOG_PACKET, "=== END PACKET ===\n");
}

void lwins_dump_discovery_raw_packet (unsigned char *raw_packet, size_t len, struct sockaddr_in *client)
{
    u16 tmpu16;
    struct nbns_discover_packet packet;
    
    LOG (LOG_PACKET, "=== DISCOVERY PACKET FROM %s:%u SIZE: %zd ===\n", inet_ntoa (client->sin_addr), ntohs(client->sin_port), len);
    
    memset (&packet, 0, sizeof (packet));
    if (parse_raw_nbns_discover_packet (raw_packet, len, &packet))
    {
        LOG (LOG_PACKET, "- HEADER\n");
        LOG (LOG_PACKET, "--- op_code: %u\n", packet.op_code);
        LOG (LOG_PACKET, "- ADDRESSES: %u\n", packet.addresses_cnt);
        for (tmpu16 = 0; tmpu16 < packet.addresses_cnt; tmpu16++)
            LOG (LOG_PACKET, "-- addresses[%u]: %s\n", tmpu16, inet_ntoa (packet.addresses[tmpu16]));
    }
    else
        LOG (LOG_PACKET, "Could not parse packet\n");
    
    free_nbns_discover_packet (&packet);
    
    lwins_dump_buffer_hex (raw_packet, len);
    
    LOG (LOG_PACKET, "=== END PACKET ===\n");
}
